import pytest
from pyparsing.exceptions import ParseException

from ..expressions import ASTNode, expression


def test_schema_expressions(schema_obj):
    # Basic smoke test. All of these expressions should parse without error.
    for testexp in schema_obj.meta.expression_tests:
        expression.parse_string(testexp["expression"])


@pytest.mark.parametrize(
    "expr, disambiguated",
    [
        ("a*x**2 + b*x + c == 0", "((((a * (x ** 2)) + (b * x)) + c) == 0)"),
        ("a.b.c[0][1][2].d(3, 4)", "(((a.b).c)[0][1][2].d)(3, 4)"),
        (
            "nifti_header.pixdim[4] == sidecar.RepetitionTime",
            "((nifti_header.pixdim)[4] == (sidecar.RepetitionTime))",
        ),
        (
            """x.y * 10 ** (-3 * index(["a", "b", "c"], w.z))""",
            """((x.y) * (10 ** (-3 * index(['"a"', '"b"', '"c"'], (w.z)))))""",
        ),
        ("a && b && c || d || e && f", "((a && (b && c)) || (d || (e && f)))"),
        ("! a && b || ! c == d || e", "(((!a) && b) || ((!(c == d)) || e))"),
    ],
)
def test_expression_associations(expr, disambiguated):
    # The goal of this test is to use the str() function on the AST to
    # reproduce the original expression, but with parentheses demonstrating
    # associativity. For example `a.b.c` could be parsed `(a.b).c` or `a.(b.c)`.
    parse_results = expression.parse_string(expr)
    assert len(parse_results) == 1
    ast = parse_results[0]
    assert str(ast) == disambiguated


def test_selectors(schema_obj):
    # Selectors happen at the second level in a file
    # For directories of files, need to look three down
    for rules, level in (
        (schema_obj.rules.checks, 3),
        (schema_obj.rules.dataset_metadata, 2),
        (schema_obj.rules.sidecars, 3),
        (schema_obj.rules.tabular_data, 3),
        (schema_obj.rules.errors, 2),
    ):
        keys = (key for key in rules.keys(level=level) if key.endswith("selectors"))
        for key in keys:
            for selector in rules[key]:
                ast = expression.parse_string(selector)[0]
                assert isinstance(ast, ASTNode)


def test_checks(schema_obj):
    checks = (
        expr
        for key, value in schema_obj.rules.checks.items(level=3)
        if key.endswith("checks")
        for expr in value
    )
    for check in checks:
        ast = expression.parse_string(check)[0]
        assert isinstance(ast, ASTNode)


@pytest.mark.parametrize(
    "expr",
    (
        "func(x, y",
        "lhs == ",
        "assignment = not.a.thing",
    ),
)
def test_expected_failures(expr):
    with pytest.raises(ParseException):
        expression.parse_string(expr)



"""A Python package for working with the BIDS schema.

.. autodata:: __version__
.. autodata:: __bids_version__
"""

try:  # Prefer backport to leave consistency to dependency spec
    from importlib_resources import files
except ImportError:
    from importlib.resources import files  # type: ignore

version_file = files("bidsschematools.data") / "schema" / "SCHEMA_VERSION"
__version__ = version_file.read_text().strip()
"Schema version"

bids_version_file = files("bidsschematools.data") / "schema" / "BIDS_VERSION"
__bids_version__ = bids_version_file.read_text().strip()
"BIDS specification version"



"""Utility functions for the bids-specification schema."""

import logging

from . import data


def get_bundled_schema_path():
    """Get the path to the schema directory.

    Returns
    -------
    str
        Absolute path to the directory containing schema-related files.
    """
    return str(data.load_resource("schema"))


def get_logger(name=None):
    """Return a logger to use.

    Parameters
    ----------
    name : None or str, optional
        Name of the logger. Defaults to None.

    Returns
    -------
    logging.Logger
        logger object.
    """
    return logging.getLogger("bids-schema" + (".%s" % name if name else ""))


def set_logger_level(lgr, level):
    """Set the logger level.

    Parameters
    ----------
    lgr : logging.Logger
        logger object for which to change the level.
    level : int or str
        The desired level for the logger.
    """
    if isinstance(level, int):
        pass
    elif level.isnumeric():
        level = int(level)
    elif level.isalpha():
        level = getattr(logging, level)
    else:
        lgr.warning("Do not know how to treat loglevel %s" % level)
        return
    lgr.setLevel(level)



import logging
import os
import sys

import click

if sys.version_info < (3, 9):
    from importlib_resources import files
else:
    from importlib.resources import files


from .schema import export_schema, load_schema


@click.group()
@click.option("-v", "--verbose", count=True)
def cli(verbose):
    """BIDS Schema Tools"""
    logging.getLogger("bidsschematools").setLevel(logging.INFO - verbose * 10)


@cli.command()
@click.option("--schema")
@click.option("--output", default="-")
@click.pass_context
def export(ctx, schema, output):
    """Export BIDS schema to JSON document"""
    logger = logging.getLogger("bidsschematools")
    schema = load_schema(schema)
    text = export_schema(schema)
    if output == "-":
        logger.debug("Writing to stdout")
        print(text)
    else:
        output = os.path.abspath(output)
        logger.debug(f"Writing to {output}")
        with open(output, "w") as fobj:
            fobj.write(text)


@cli.command()
@click.option("--output", default="-")
@click.pass_context
def export_metaschema(ctx, output):
    """Export BIDS schema to JSON document"""
    metaschema = files("bidsschematools.data").joinpath("metaschema.json").read_text()
    if output == "-":
        print(metaschema, end="")
    else:
        output = os.path.abspath(output)
        with open(output, "w") as fobj:
            fobj.write(metaschema)


if __name__ == "__main__":
    cli()



"""Module containing data files, including the schema source files

.. autofunction:: load_resource
"""

import atexit
import os
from contextlib import ExitStack
from functools import cached_property
from pathlib import Path
from types import ModuleType
from typing import Union

try:
    from functools import cache
except ImportError:  # PY38
    from functools import lru_cache as cache

try:  # Prefer backport to leave consistency to dependency spec
    from importlib_resources import as_file, files
except ImportError:
    from importlib.resources import as_file, files  # type: ignore

__all__ = ["load_resource"]


class Loader:
    """A loader for package files relative to a module

    This class wraps :mod:`importlib.resources` to provide a getter
    function with an interpreter-lifetime scope. For typical packages
    it simply passes through filesystem paths as :class:`~pathlib.Path`
    objects. For zipped distributions, it will unpack the files into
    a temporary directory that is cleaned up on interpreter exit.

    This loader accepts a fully-qualified module name or a module
    object.

    Expected usage::

        '''Data package

        .. autofunction:: load_data
        '''

        from bidsschematools.data import Loader

        load_data = Loader(__package__)

    :class:`~Loader` objects implement the :func:`callable` interface
    and generate a docstring, and are intended to be treated and documented
    as functions.
    """

    def __init__(self, anchor: Union[str, ModuleType]):
        self._anchor = anchor
        self.files = files(anchor)
        self.exit_stack = ExitStack()
        atexit.register(self.exit_stack.close)
        # Allow class to have a different docstring from instances
        self.__doc__ = self._doc

    @cached_property
    def _doc(self):
        """Construct docstring for instances

        Lists the public top-level paths inside the location, where
        non-public means has a `.` or `_` prefix or is a 'tests'
        directory.
        """
        top_level = sorted(
            os.path.relpath(p, self.files) + "/"[: p.is_dir()]
            for p in self.files.iterdir()
            if p.name[0] not in (".", "_") and p.name != "tests"
        )
        doclines = [
            f"Load package files relative to ``{self._anchor}``.",
            "",
            "This package contains the following (top-level) files/directories:",
            "",
            *(f"* ``{path}``" for path in top_level),
        ]

        return "\n".join(doclines)

    @cache
    def __call__(self, *segments) -> Path:
        """Ensure data is available as a :class:`~pathlib.Path`.

        Any temporary files that are created remain available throughout
        the duration of the program, and are deleted when Python exits.

        Results are cached so that multiple calls do not unpack the same
        data multiple times, but the cache is sensitive to the specific
        argument(s) passed.
        """
        return self.exit_stack.enter_context(as_file(self.files.joinpath(*segments)))


load_resource = Loader(__package__)
